"""Console script for csverve."""
import os

import click
import csverve.api as api
import yaml


@click.group()
def cli():
    pass


@cli.command()
@click.option('--in_f', multiple=True, required=True, help='CSV file path, allows multiple paths.')
@click.option('--out_f', required=True, help='Path of resulting merged CSV.')
@click.option('--how', required=True, help='How to join CSVs.')
@click.option('--on', multiple=True, required=False, help='Column to join CSVs on, allowes multiple.')
@click.option('--skip_header', is_flag=True, default=False, help='Writer header to resulting CSV.')
def merge(
        in_f,
        out_f,
        how,
        on,
        skip_header,
):
    on = list(on)
    api.merge_csv(
        list(in_f),
        out_f,
        how,
        on,
        skip_header,
    )


@cli.command()
@click.option('--in_f', required=True, help='CSV file path. Expects YAML w/ the same path as CSV with .yaml extension.')
@click.option('--out_f', required=True, help='Path of resulting merged CSV.')
@click.option('--dtypes', required=True, help='dtypes yaml file')
@click.option('--skip_header', is_flag=True, default=False, help='Writer header to resulting CSV.')
def rewrite(
        in_f,
        out_f,
        skip_header,
        dtypes
):
    assert os.path.exists(in_f)

    with open(dtypes, 'rt') as reader:
        dtypes_data = yaml.load(reader)

    api.rewrite_csv_file(
        in_f,
        out_f,
        skip_header,
        dtypes_data
    )


@cli.command()
@click.option('--in_f', required=True, help='CSV file path. Expects YAML w/ the same path as CSV with .yaml extension.')
@click.option('--out_f', required=True, help='Path of resulting merged CSV.')
@click.option('--skip_header', is_flag=True, default=False, help='Writer header to resulting CSV.')
def remove_duplicates(
        in_f,
        out_f,
        skip_header
):
    assert os.path.exists(in_f)

    api.remove_duplicates(
        in_f,
        out_f,
        skip_header
    )


@cli.command()
@click.option('--in_f', multiple=True, required=True, help='CSV file path, allows multiple paths.')
@click.option('--out_f', required=True, help='Path of resulting merged CSV.')
@click.option('--skip_header', is_flag=True, default=False, help='Writer header to resulting CSV.')
@click.option('--drop_duplicates', is_flag=True, default=False, help='remove duplicate rows')
def concat(
        in_f,
        out_f,
        skip_header,
        drop_duplicates
):
    api.concatenate_csv(
        list(in_f),
        out_f,
        skip_header,
        drop_duplicates
    )


@cli.command()
@click.option('--in_f', required=True, help='CSV file path, allows multiple paths.')
@click.option('--out_f', required=True, help='Path of resulting merged CSV.')
@click.option('--col_name', required=True, help='Column name to be added.')
@click.option('--col_val', required=True, help='Column value to be added (one value for all).')
@click.option('--col_dtype', required=True, help='Column pandas dtype.')
@click.option('--skip_header', is_flag=True, default=False, help='Writer header to resulting CSV.')
def annotate(
        in_f,
        out_f,
        col_name,
        col_val,
        col_dtype,
        skip_header,
):
    api.simple_annotate_csv(
        in_f,
        out_f,
        col_name,
        col_val,
        col_dtype,
        skip_header,
    )


if __name__ == "__main__":
    cli()



# type: ignore
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


from csverve.api import rewrite_csv_file
from csverve.api import merge_csv
from csverve.api import concatenate_csv
from csverve.api import annotate_csv
from csverve.api import simple_annotate_csv
from csverve.api import add_col_from_dict
from csverve.api import write_dataframe_to_csv_and_yaml
from csverve.api import read_csv
from csverve.api import get_columns
from csverve.api import get_dtypes
from csverve.api import remove_duplicates


import os
import random
import string

import csverve.api as api
import pandas as pd
import yaml as y
from csverve.core.csverve_input import CsverveInput
from csverve.errors import CsverveParseError


class TestInputs:

    def write_dfs(self, tmpdir, dfs, dtypes, skip_header=False):
        n_dfs = len(dfs)
        names = [os.path.join(tmpdir, str(i) + ".csv.gz") for i in range(n_dfs)]

        assert len({n_dfs, len(dtypes)}) == 1

        for i in range(n_dfs):
            api.write_dataframe_to_csv_and_yaml(dfs[i], names[i],
                                                dtypes[i], skip_header)
        return names

    def make_test_df(self, dtypes, length):

        df_dict = {}
        for col, dtype in dtypes.items():
            df_dict[col] = self.simulate_col(dtype, length)

        df = pd.DataFrame(df_dict, columns=dtypes.keys())
        df = df.astype(dtypes)

        return df

    def make_test_dfs(self, dtypes, length):

        n_dfs = len(dtypes)
        dfs = [self.make_test_df(dtypes[i], length) for i in range(n_dfs)]

        return dfs

    def simulate_col(self, dtype, length):
        if dtype == "str":
            return self._str_list(length)
        if dtype == "int":
            return self._rand_int_col(length)
        if dtype == "float":
            return self._rand_float_col(length)
        if dtype == "bool":
            return self._rand_bool_col(length)

    def _rand_float_col(self, n):
        return [random.uniform(0, 100) for _ in range(n)]

    def _rand_int_col(self, n, scale=1):
        return list(range(n)) * scale
        # return [random.randint(0, 10) for _ in range(n)]

    def _rand_bool_col(self, n):
        return [random.choice([True, False]) for _ in range(n)]

    def _str_list(self, n, must_have="", count=0):
        s = random.sample(string.ascii_uppercase, k=n)
        if count:
            s = [l + str(count) for l in s]
        if must_have != "":
            s.append(must_have)
        return s


class TestValidationHelpers:

    def _raises_correct_error(self, function, *args,
                              expected_error=CsverveParseError,
                              **kwargs):
        raised = False
        try:
            function(*args, **kwargs)
        except Exception as e:
            if type(e) == expected_error:
                raised = True
            else:
                print("raised wrong error: raised: {}, expected: {}"
                      .format(type(e), expected_error))
        finally:
            return raised

    def dfs_exact_match(self, data, reference):

        if isinstance(data, str):
            data = CsverveInput(data).read_csv()
        if isinstance(reference, str):
            reference = CsverveInput(reference).read_csv()

        if set(data.columns) != set(reference.columns):
            return False

        # read csv doesnt precisely read floats
        data = data.round(5)
        reference = reference.round(5)

        return all([reference[col].equals(data[col]) for col in reference.columns])


class AnnotationHelpers(TestInputs, TestValidationHelpers):
    """
    helper functions for testing the annotate_csv csverve function
    """

    def sim_ann_input(self, annotate_col, annotate_col_name, dtypes):
        """
        make annotation dict for test annotate_csv.
        :param annotate_col: column to create annotation on
        :param dtypes: dtypes for annotation dict
        """

        annotations = []

        for cell_id in annotate_col:
            outdict = {col: self.simulate_col(dtype, 1)[0] for col, dtype in dtypes.items()}
            outdict[annotate_col_name] = cell_id
            annotations.append(outdict)

        annotations = pd.DataFrame(annotations)

        return annotations

    def make_ann_test_inputs(self, temp, length, dtypes, ann_dtypes,
                             skip_header=False, on="cell_id"):
        """
        make inputs to test annotate_csv
        :param temp: tempdir to test in
        :param length: length of test dfs
        :param dtypes: dtypes of test dfs
        :param ann_dtypes: dtypes of test annotation dict
        :param skip_header: T/F write header post-annotation
        :param on: col to annotate on
        """
        df = self.make_test_dfs([dtypes], length)
        csv = self.write_dfs(temp, df, [dtypes], skip_header)

        df = df[0]
        csv = csv[0]

        annotation_input = self.sim_ann_input(df[on].tolist(), on, ann_dtypes)

        return csv, annotation_input

    def base_annotation_test(self, temp, length, dtypes, ann_dtypes, skip_header=False,
                             on="cell_id"):
        """
        base test of annotate_csv
        :param temp: tempdir to test in
        :param length: length of test dfs
        :param dtypes: dtypes of test dfs
        :param ann_dtypes: dtypes of test annotation dict
        :param skip_header: T/F write header post-annotation
        :param on: col to annotate on:
        """

        csv, annotation = self.make_ann_test_inputs(temp, length,
                                                    dtypes, ann_dtypes,
                                                    skip_header=skip_header,
                                                    on=on)

        annotated = os.path.join(temp, "annotated.csv.gz")

        api.annotate_csv(csv, annotation, annotated, ann_dtypes,
                         skip_header=skip_header, on=on)

        return csv, annotation, annotated

    def validate_annotation_test(self, csv, annotation, annotated, on):
        """
        test that annotate_csv produced correct result in tests
        :param csv: csv that was annotated
        :param annotation: input annotation dict
        :param annotated: csv post-annotation
        :param on: col on which csv was annotated
        :return:
        """
        if isinstance(csv, str):
            csv = CsverveInput(csv).read_csv()

        compare = csv.merge(annotation, how="outer", on=on)

        return os.path.exists(annotated) \
               and self.dfs_exact_match(compare, annotated)


class ConcatHelpers(TestInputs, TestValidationHelpers):
    """
    helpers for testing concat functions
    """

    def base_test_concat(self, length, dtypes, write=False, dir=None,
                         get_ref=False, skip_header=False):
        """
        base concatenation test; make dfs, write them,
        get a "reference" concat output
        :param length: number of rows in df; use length not n_rows so as to not use fixture
        :param dtypes: list of dtypes to use in creation of test dfs
        :param write: T/F :write out dfs to csvs
        :param dir: directory to write to if writing
        :param get_ref: T/F: get excpected output from concatenation using
        pandas built-ins
        :param skip_header: T/F: write header when writing out to csvs
        :return: dfs and or csvs and or expected output
        """

        dfs = self.make_test_dfs(dtypes, length)

        outs = [dfs]

        if write:
            csvs = self.write_dfs(dir, dfs, dtypes, skip_header)
            outs.append(csvs)

        if get_ref:
            ref = pd.concat(dfs, ignore_index=True)
            outs.append(ref)

        return tuple(outs)


class WriteHelpers(TestInputs, TestValidationHelpers):
    """
    helpers class for testing of csverve writing
    """

    def write_file_successful(self, df, csv, skipped_header=False):
        head = "infer"
        if skipped_header:
            head = None
        csv = pd.read_csv(csv, sep=",", header=head)

        if skipped_header:
            csv.columns = df.columns

        if not df.equals(csv):
            return False
        return True

    def metadata_write_successful(self, dtypes, yaml_file):
        """
        validate writing of dtypes to yaml
        :param dtypes: written dtypes
        :param yaml_file: output .yaml
        :return:
        """
        yaml_loader = y.load(open(yaml_file), Loader=y.FullLoader)["columns"]
        yaml = {}

        for dtype in yaml_loader:
            yaml[dtype["name"]] = dtype["dtype"]

        if yaml != dtypes:
            return False
        return True


class MergeHelpers(TestInputs, TestValidationHelpers):
    """
    helpers class for testing of merging
    """

    def make_mergeable_test_dfs(self, dtypes, shared, length):
        """
        make dfs that can be merged
        :param dtypes: list of dtypes corresponding to test dfs
        :param shared: columns shared by dfs, should all be in dtypes
        :param length: length of test dfs
        :return:
        """

        for dtype_set in dtypes:
            if not set(shared).issubset(set(dtype_set.keys())):
                raise Exception(shared, dtype_set, dtype_set.keys())

            assert set(shared).issubset(set(dtype_set.keys()))

        dfs = self.make_test_dfs(dtypes, length)

        shared_values = dfs[0][shared]

        for df in dfs:
            df.update(shared_values)

        return dfs

    def base_merge_test(self, length, how, on, suffixes, dtypes, write=False,
                        get_ref=True, dir=None, skip_header=False):
        """
        base test for all merge fnctions
        :param length: length of test dfs
        :param how: how to merge (i.e. direction)
        :param on: col(s) to merge on
        :param suffixes: suffixes for merged df (see pandas docs)
        :param dtypes: list of dtypes for creation of test dfs
        :param write: T/F write out test dfs before merging
        :param get_ref: get expected merge output
        :param dir: temp dir to test in
        :param skip_header: T/F write header to csv (before and after testing)
        """

        dfs = self.make_mergeable_test_dfs(dtypes, on, length)

        outs = [dfs]
        if not write and not get_ref:
            return dfs

        if write:
            csvs = self.write_dfs(dir, dfs, dtypes, skip_header)
            outs.append(csvs)

        if get_ref:
            ref = dfs[0].merge(dfs[1], how=how, on=on, suffixes=suffixes)
            outs.append(ref)

        return tuple(outs)



from csverve.core.csverve_input import CsverveInput
from csverve.core.csverve_output import CsverveOutput
from csverve.core.csverve_output_data_frame import CsverveOutputDataFrame
from csverve.core.csverve_output_file_stream import CsverveOutputFileStream
from csverve.core.irregular_csv_input import IrregularCsverveInput



from typing import List, Dict, Any

import yaml
from csverve import utils
from csverve.errors import CsverveWriterError


class CsverveOutput(object):
    def __init__(
            self,
            filepath: str,
            dtypes: Dict[str, str],
            columns: List[str],
            skip_header: bool = False,
            na_rep: str = 'NaN',
            sep: str = ',',
    ) -> None:
        self.filepath: str = filepath
        self._verify_input()

        self.skip_header: bool = skip_header
        self.dtypes: Dict[str, str] = dtypes
        self.na_rep: str = na_rep
        self.sep: str = sep
        self._convert_dtypes()

        self.columns = columns

    def _convert_dtypes(self) -> None:
        self.dtypes = {col: utils.pandas_to_std_types(dtype) for col, dtype in self.dtypes.items()}

    @property
    def yaml_file(self) -> str:
        """
        Append '.yaml' to CSV path.

        @return: YAML metadata path.
        """
        return self.filepath + '.yaml'

    def _verify_input(self):
        """
        Verify gzip status and check for yaml

        @return:
        """
        if not self.filepath.endswith('.gz'):
            raise CsverveWriterError('output must be gzipped')

    def write_yaml(self) -> None:
        """
        Write .yaml file.

        @return:
        """

        yaml_columns = [{'name': col, 'dtype': self.dtypes[col]} for col in self.columns]

        yamldata: Dict[str, Any] = {
            'header': (not self.skip_header),
            'sep': self.sep,
            'columns': yaml_columns
        }

        with open(self.yaml_file, 'wt') as f:
            yaml.safe_dump(yamldata, f, default_flow_style=False)



import gzip
import shutil
from typing import List, Dict, TextIO, Any

from csverve.core import CsverveOutput
from csverve.errors import CsverveInputError


class CsverveOutputFileStream(CsverveOutput):
    def __init__(
            self,
            filepath: str,
            dtypes: Dict[str, str],
            columns: List[str],
            skip_header: bool = False,
            na_rep: str = 'NaN',
            sep: str = ',',
    ) -> None:
        """
        CSV file and all related metadata.

        @param filepath: CSV file path.
        @param dtypes: Dictionary of pandas dtypes, where key = column name, value = dtype.
        @param header: boolean, True = write header, False = don't write header.
        @param na_rep: replace NaN with this value.
        @param columns: List of column names.
        """

        super().__init__(
            filepath, dtypes, columns,
            skip_header=skip_header, na_rep=na_rep, sep=sep
        )

    def _write_header_to_file(self, writer: TextIO) -> None:
        """
        Write header.
        @param writer: TextIO.
        @return:
        """
        assert self.columns
        header: str = ','.join(self.columns)
        header = header + '\n'
        writer.write(header)

    def write_data_streams(self, csvfiles: List[str]) -> None:
        """
        Write data streams.
        @param csvfiles: List of CSV files paths.
        @return:
        """
        assert self.columns
        assert self.dtypes
        with gzip.open(self.filepath, 'wt') as writer:

            if not self.skip_header:
                self._write_header_to_file(writer)

            for csvfile in csvfiles:
                with gzip.open(csvfile, 'rt') as data_stream:
                    shutil.copyfileobj(
                        data_stream, writer, length=16 * 1024 * 1024
                    )

        self.write_yaml()

    @staticmethod
    def _file_type(filepath) -> str:
        if filepath.endswith('gz'):
            return 'gzip'
        elif filepath.endswith('csv'):
            return 'plain-text'
        else:
            raise CsverveInputError('Unsupported file type: {}'.format(filepath))

    def rewrite_csv(self, csvfile: str) -> None:
        """
        Rewrite CSV.
        @param csvfile: Filepath of CSV file.
        @return:
        """

        assert self.columns
        assert self.dtypes

        filetype = self._file_type(csvfile)

        opener: Any = gzip.open if filetype == 'gzip' else open

        with gzip.open(self.filepath, 'wt') as writer:
            with opener(csvfile, 'rt') as data_stream:
                shutil.copyfileobj(
                    data_stream, writer, length=16 * 1024 * 1024
                )

        self.write_yaml()



class CsverveMergeDtypesEmptyMergeSet(Exception):
    pass


class CsverveMergeCommonColException(Exception):
    pass


class DtypesMergeException(Exception):
    pass


class CsverveConcatException(Exception):
    pass


class CsverveAnnotateError(Exception):
    pass


class CsverveMergeException(Exception):
    pass


class CsverveMergeColumnMismatchException(Exception):
    pass


class CsverveParseError(Exception):
    pass


class CsverveInputError(Exception):
    pass


class CsverveWriterError(Exception):
    pass


class CsverveDtypeError(Exception):
    pass



from typing import Any
from typing import List, Dict, Union

import pandas as pd  # type: ignore
from csverve.errors import CsverveDtypeError
from csverve.errors import CsverveMergeColumnMismatchException
from csverve.errors import CsverveMergeCommonColException
from csverve.errors import CsverveMergeDtypesEmptyMergeSet
from csverve.errors import CsverveMergeException
from csverve.errors import DtypesMergeException


def pandas_to_std_types(dtype: Any) -> str:
    std_dict = {
        "bool": "bool",
        "int64": "int",
        "int": "int",
        "Int64": "int",
        "float64": "float",
        "float": "float",
        "object": "str",
        "str": "str",
        "category": "category",
    }

    if not isinstance(dtype, str):
        if hasattr(dtype, '__name__'):
            dtype = dtype.__name__
        else:
            raise CsverveDtypeError('Unable to process dtype {}'.format(dtype))

    if dtype not in std_dict:
        raise CsverveDtypeError('Unable to process dtype {}'.format(dtype))

    return std_dict[dtype]


def merge_dtypes(dtypes_all: List[Dict[str, str]]) -> Dict[str, str]:
    """
    Merge pandas dtypes.

    @param dtypes_all: List of dtypes dictionaries, where key = column name, value = pandas dtype.
    @return: Merged dtypes dictionary.
    """

    if not dtypes_all:
        raise CsverveMergeDtypesEmptyMergeSet("must provide dtypes to merge")

    merged_dtypes: Dict[str, str] = {}

    for dtypes in dtypes_all:
        for k, v in dtypes.items():
            if k in merged_dtypes:
                if merged_dtypes[k] != v:
                    raise DtypesMergeException("dtypes not mergeable")
            else:
                merged_dtypes[k] = v

    return merged_dtypes


def _validate_merge_cols(frames: List[pd.DataFrame], on: Union[List[str], str]) -> None:
    """
    Make sure frames look good, raise relevant exceptions.

    @param frames: list of pandas DataFrames to merge
    @param on: list of common columns in frames on which to merge
    @return:
    """

    if not on:
        raise CsverveMergeException("unable to merge if given nothing to merge on")

    # check that columns to be merged have identical values
    standard = frames[0][on]
    standard = standard.sort_values(on).reset_index(drop=True)
    for frame in frames:
        comp_df = frame[on].sort_values(on).reset_index(drop=True)
        if not pd.concat([standard, comp_df]).drop_duplicates(keep=False).empty:
            raise CsverveMergeColumnMismatchException("columns on which to merge must be identical")

    # check that columns to be merged have same dtypes
    for shared_col in on:
        if len(set([frame[shared_col].dtype.name for frame in frames])) != 1:
            raise CsverveMergeColumnMismatchException("columns on which to merge must have same dtypes")

    common_cols = set.intersection(*[set(frame.columns) for frame in frames])
    cols_to_check = list(common_cols - set(on))

    for frame1, frame2 in zip(frames[:-1], frames[1:]):
        if not frame1[cols_to_check].equals(frame2[cols_to_check]):
            raise CsverveMergeCommonColException("non-merged common cols must be identical")


def merge_frames(frames: List[pd.DataFrame], how: str, on: List[str], lenient: bool = False) -> pd.DataFrame:
    """
    Takes in a list of pandas DataFrames, and merges into a single DataFrame.
    #TODO: add handling if empty list is given

    @param frames: List of pandas DataFrames.
    @param how: How to join DataFrames (inner, outer, left, right).
    @param on: Column(s) to join on, comma separated if multiple.
    @param lenient: allow merge of mismatched csvs
    @return: merged pandas DataFrame.
    """
    if isinstance(on, str):
        on = [on]

    if lenient is False:
        _validate_merge_cols(frames, on)

    if len(frames) == 1:
        return frames[0]
    else:
        left: pd.DataFrame = frames[0]
        right: pd.DataFrame = frames[1]
        cols_to_use: List[str] = list(right.columns.difference(left.columns))
        cols_to_use += on
        cols_to_use = list(set(cols_to_use))

        merged_frame: pd.DataFrame = pd.merge(
            left, right[cols_to_use], how=how, on=on,
        )

        for i, frame in enumerate(frames[2:]):
            merged_frame = pd.merge(
                merged_frame, frame, how=how, on=on,
            )
        return merged_frame
